from pyxk.lazy_loader import LazyLoader
from pyxk.m3u8._download import _AsyncDownload

os = LazyLoader("os", globals())
shlex = LazyLoader("shlex", globals())
warnings = LazyLoader("warnings", globals())
threading = LazyLoader("threading", globals())
subprocess = LazyLoader("subprocess", globals())

rich = LazyLoader("rich", globals())
_m3u8 = LazyLoader("_m3u8", globals(), "m3u8")
panel = LazyLoader("panel", globals(), "rich.panel")
columns = LazyLoader("columns", globals(), "rich.columns")
progress = LazyLoader("progress", globals(), "rich.progress")

utils = LazyLoader("utils", globals(), "pyxk.utils")
requests = LazyLoader("requests", globals(), "pyxk.requests")



class M3U8():
    """m3u8 下载器"""

    default_limit = 8
    default_filename = "index.mp4"

    def __init__(self):

        self._headers = {}
        self._verify  = True
        self._file = None
        self._temp = None
        self._limit = self.default_limit
        self._rerequest = False
        self._reserve_segments = False
        self._delete_m3u8file = False
        self._session = requests.Session()


    def _get_m3u8_content(self, uri, file, encoding="UTF-8", *, m3u8key=False):
        """获取 m3u8 文件内容"""
        file = os.path.join(self._temp, utils.hash256(file))
        file = file + ".key" if m3u8key else file + ".m3u8"

        if self._rerequest or not os.path.isfile(file):
            response = self._session.get(url=uri, headers=self._headers, verify=self._verify)
            return getattr(response, "content") if m3u8key else getattr(response, "text")

        mode, encoding = "rb" if m3u8key else "r", None if m3u8key else encoding
        with open(file, mode, encoding=encoding) as fileobj:
            content = fileobj.read()
        return content


    def _sava_m3u8_content(self, file, content, encoding="UTF-8", *, m3u8key=False):
        """保存 m3u8 文件内容"""
        if not isinstance(file, str) or not file:
            return
        file = os.path.join(self._temp, utils.hash256(file))
        file = file + ".key" if m3u8key else file + ".m3u8"

        if self._rerequest or not os.path.isfile(file):
            mode, encoding = "wb" if m3u8key else "w", None if m3u8key else encoding
            with utils.open(file, mode, encoding=encoding) as fileobj:
                fileobj.write(content)


    def _file_abspath(self, file, /):
        """文件绝对路径"""
        if not isinstance(file, str) or not file:
            file = self.default_filename
        file = "-".join(os.path.normpath(os.path.abspath(file)).strip().split())
        file = utils.rename_file(file, suffix="mp4")
        self._file = file[0]
        self._temp = os.path.join(file[1], file[2].removesuffix("mp4") + "temp")


    def _initialization(
        self, file=None, headers=None, verify=True, rerequest=False, delete=False, reserve=False, limit=8
    ):
        self._file_abspath(file)
        self._headers = dict(headers) if headers else {}
        self._verify = bool(verify)
        self._rerequest = bool(rerequest)
        self._delete_m3u8file = bool(delete)
        self._reserve_segments = bool(reserve)
        if not isinstance(limit, int) or limit <= 0:
            limit = self.default_limit
        self._limit = limit


    def load(
        self, uri, **kwargs
    ):
        """解析 m3u8 链接"""
        self._initialization(**kwargs)
        content = self._get_m3u8_content(uri=uri, file=uri)
        return self._start(content=content, uri=uri)


    def loads(
        self, content, uri=None, **kwargs
    ):
        """解析 m3u8 链接"""
        self._initialization(**kwargs)
        if content.startswith("#EXTM3U"):
            return self._start(content=content, uri=uri)

        _m3u8file = os.path.normpath(os.path.abspath(content))
        if os.path.isfile(_m3u8file):
            with open(_m3u8file, "r", encoding="utf-8") as fileobj:
                content = fileobj.read()
        return self._start(content=content, uri=uri)


    def _start(self, content, uri=None):
        """解析 m3u8 入口"""
        # playlists
        m3u8_obj, variant_uri = self._parse_playlist(
            m3u8_obj=_m3u8.loads(content=content, uri=uri), uri=uri
        )
        # m3u8 keys
        m3u8_keys = self._parse_m3u8_keys(m3u8_obj=m3u8_obj)
        # segments
        segments, duration = self._parse_segments(
            m3u8_obj=m3u8_obj, uri=variant_uri, m3u8_keys=m3u8_keys
        )
        self._display(uri=uri, maximum=len(segments), duration=duration, encryption=bool(m3u8_keys))

        # 异步下载
        download = _AsyncDownload(store=self._temp)
        download_result = download.start(segments=segments, m3u8_keys=m3u8_keys)

        # 合并 segments
        merge_result = self._merge_segments(files=download_result)
        # 删除 segments
        if merge_result and not self._reserve_segments:
            for segment in download_result:
                os.remove(segment)
            if not os.listdir(download.store):
                os.rmdir(download.store)
        # 删除 m3u8 文件
        if self._delete_m3u8file and os.path.isdir(self._temp):
            for file in os.listdir(self._temp):
                file = os.path.join(self._temp, file)
                if not os.path.isfile(file):
                    continue
                os.remove(file)
            if not os.listdir(self._temp):
                os.rmdir(self._temp)


    def _merge_segments(self, files):
        if not files:
            return False

        # 生成filelists文件
        filelists, filesize = os.path.join(self._temp, "filelist.txt"), 0
        with utils.open(filelists, "w", encoding="utf-8") as fileobj:
            for file in files:
                fileobj.write(f"file '{file}'\n")
                filesize += os.path.getsize(file) - 16400

        # ffmpeg合并代码
        merge_code = shlex.split(
            f"ffmpeg -loglevel quiet -f concat -safe 0 -i {filelists} -c copy {self._file} -y"
        )
        merge_complete = False

        # 合并函数
        def merge():
            try:
                subprocess.run(args=merge_code, check=True)
                os.remove(filelists)
            except FileNotFoundError as error:
                reason = getattr(error, "filename", None)
                if reason != "ffmpeg":
                    raise
                warnings.warn("没有ffmpeg, 调用失败")
            finally:
                nonlocal merge_complete
                merge_complete = True

        # 合并进度条函数
        def merge_progress():
            last_filesize = 0
            get_filesize = lambda file: os.path.getsize(file) if os.path.isfile(file) else 0
            # 进度条
            with progress.Progress(
                *(
                    progress.SpinnerColumn("line"),
                    progress.DownloadColumn(),
                    progress.BarColumn(finished_style="green"),
                    progress.TaskProgressColumn("[progress.percentage]{task.percentage:>6.2f}%"),
                    progress.TimeElapsedColumn()
                ),
                transient=True
            ) as progressing:

                progress_task = progressing.add_task(description=None,total=filesize)
                while True:
                    current_filesize = get_filesize(self._file)
                    progressing.update(progress_task, advance=current_filesize-last_filesize)
                    last_filesize = current_filesize
                    # 控制进度条退出
                    if merge_complete:
                        progressing.update(progress_task, advance=abs(filesize-last_filesize))
                        break

        task1 = threading.Thread(target=merge)
        task2 = threading.Thread(target=merge_progress)
        task1.start()
        task2.start()
        task1.join()
        task2.join()
        return os.path.isfile(self._file)


    def _parse_playlist(self, m3u8_obj, uri=None):
        if not m3u8_obj.is_variant:
            return m3u8_obj, uri

        def sort_playlist(playlist):
            playlist.uri = playlist.absolute_uri
            return playlist.stream_info.bandwidth
        playlists = sorted(m3u8_obj.playlists, key=sort_playlist)

        # 保存playlists
        self._sava_m3u8_content(file=uri, content=m3u8_obj.dumps())

        variant_uri = playlists[-1].uri
        content = self._get_m3u8_content(uri=variant_uri, file=variant_uri)

        return self._parse_playlist(
            m3u8_obj=_m3u8.loads(
                content=content, uri=variant_uri
            ),
            uri=variant_uri
        )


    def _parse_m3u8_keys(self, m3u8_obj):
        m3u8_keys = {}
        for index, key in enumerate(m3u8_obj.keys):
            if not key:
                continue
            key.uri = key.absolute_uri
            secret  = self._get_m3u8_content(uri=key.uri, file=key.uri, m3u8key=True)
            # 保存密钥
            self._sava_m3u8_content(file=key.uri, content=secret, m3u8key=True)
            m3u8_keys[index] = {
                "cipher": {
                    "key": secret,
                    "iv": key.iv.removeprefix("0x")[:16] if key.iv else secret[:16]
                },
                "uri": key.uri,
                "method": key.method
            }
        return m3u8_keys


    def _parse_segments(self, m3u8_obj, uri=None, m3u8_keys=None):
        segments, duration = {}, 0
        for index, segment in enumerate(m3u8_obj.segments):
            segment.uri, key_flags = segment.absolute_uri, None
            if segment.key and m3u8_keys:
                for key, val in m3u8_keys.items():
                    if segment.key.method == val["method"] and segment.key.uri == val["uri"]:
                        key_flags = key
                        break
            segments[index] = {"uri": segment.uri, "key": key_flags}
            duration += segment.duration

        if m3u8_obj.is_endlist:
            self._sava_m3u8_content(file=uri, content=m3u8_obj.dumps())
        return segments, duration


    def _display(self, uri, maximum, duration, encryption):
        visual_data = [
            f"Limit: {self._limit}",
            f"Maximum: {maximum}",
            f"PlayTime: {utils.human_playtime_pr(duration)}",
            f"Encryption: {encryption}",
            f"FileName: [magenta]{os.path.basename(self._file)}[/]",
            f"FilePath: {os.path.dirname(self._file)}"
        ]
        visual_data = [f"{i+1} {x}" for i, x in enumerate(visual_data)]
        if uri:
            visual_uri = panel.Panel(columns.Columns([uri], expand=True, align="center"))
            visual_data.insert(0, visual_uri)
        visual_data.append(
            "[green b]Parsing success ![/]" if maximum \
            else "[red b]Parsing failure ![/]"
        )
        visual_data = columns.Columns(visual_data, expand=True,)
        visual_data = panel.Panel(visual_data, border_style="yellow")
        rich.print(visual_data)



"""
AES加解密 数据初始化
"""
from pyxk.lazy_loader import LazyLoader

AES = LazyLoader("AES", globals(), "Crypto.Cipher.AES")



# 目前支持的模式
MODES = {
    "ECB": 1,  "CBC": 2, "CFB": 3,
    "OFB": 5,  "CTR": 6, "OPENPGP": 7,
    "EAX": 9,  "CCM": 8, "SIV": 10,
    "GCM": 11, "OCB": 12
}



class FormatData:
    """
    AES数据初始化
    """
    def __init__(
        self, key, iv=None, mode="CBC", encode="UTF-8", **kwargs
    ):
        self._key  = key
        self._mode = mode
        self._iv   = iv
        self._encode = encode
        self._state = kwargs
        self.__initialization()


    def __initialization(self):
        """
        初始化 key mode iv
        """
        self.__key_to_bytes()
        self.__mode_fmt()
        self.__iv_to_bytes()


    def __key_to_bytes(self):
        """
        key 转换为 bytes
        """
        key = self.key
        if isinstance(key, str):
            key = key.encode(self._encode)

        elif not isinstance(key, bytes):
            raise TypeError(
                "\033[31m'key' type must be 'str' or 'bytes',"
                f" not '{type(key).__name__}'\033[0m")

        # key 长度判断
        key_lenght = len(key)
        if key_lenght not in AES.key_size:
            raise ValueError(
                f"\033[31m'key' lenght must be {AES.key_size},"
                f" not '{key_lenght}'\033[0m")
        setattr(self, "_key", key)


    def __mode_fmt(self):
        """
        mode 判断
        """
        mode = self.mode
        if (
            isinstance(mode, str)
            and MODES.__contains__(mode.upper())
        ):
            mode = MODES[mode.upper()]

        if  (
            not isinstance(mode, int)
            or mode not in MODES.values()
        ):
            mode_val = list(MODES.keys())
            mode_val.extend(list(MODES.values()))
            raise TypeError(
                f"\033[31mmode must exist in the {mode_val},"
                f" not '{mode}'\033[0m")
        setattr(self, "_mode", mode)


    def __iv_to_bytes(self):
        """
        iv 转换为 bytes
        """
        iv = self.iv
        if iv is None:
            if self._mode != MODES["CBC"]:
                return
            iv = self.key[:16]

        if isinstance(iv, str):
            iv = iv.encode(self._encode)

        elif not isinstance(iv, bytes):
            raise TypeError(
                "\033[31m'iv' type must be 'str' or 'bytes',"
                f" not '{type(iv).__name__}'\033[0m")

        # iv 长度判断
        iv_lenght = len(iv)
        if iv_lenght != AES.block_size:
            raise ValueError(
                f"\033[31m'iv' lenght must be equal to '{AES.block_size}'"
                f", not '{iv_lenght}'\033[0m")
        setattr(self, "_iv", iv)


    @property
    def key(self):
        if not hasattr(self, "_key"):
            raise ValueError("\033[31m缺少'key'\033[0m")
        return getattr(self, "_key")

    @key.setter
    def key(self, value):
        setattr(self, "_key", value)
        self.__key_to_bytes()

    @property
    def mode(self):
        if not hasattr(self, "_mode"):
            setattr(self, "_mode", MODES["CBC"])
        return getattr(self, "_mode")

    @mode.setter
    def mode(self, value):
        setattr(self, "_mode", value)
        self.__mode_fmt()

    @property
    def iv(self):
        if not hasattr(self, "_iv"):
            setattr(self, "_iv", None)
        return getattr(self, "_iv")

    @iv.setter
    def iv(self, value):
        setattr(self, "_iv", value)
        self.__iv_to_bytes()



from contextlib import contextmanager
from pyxk.httpx._client import Client


def request(
    method,
    url,
    *,
    params=None,
    content=None,
    data=None,
    files=None,
    json=None,
    headers=None,
    cookies=None,
    auth=None,
    proxies=None,
    timeout=5,
    follow_redirects=True,
    verify=True,
    cert=None,
    trust_env=True,
):
    with Client(
        cookies=cookies,
        proxies=proxies,
        cert=cert,
        verify=verify,
        timeout=timeout,
        trust_env=trust_env,
    ) as client:
        return client.request(
            method=method,
            url=url,
            content=content,
            data=data,
            files=files,
            json=json,
            params=params,
            headers=headers,
            auth=auth,
            follow_redirects=follow_redirects,
        )


def get(
    url,
    *,
    params=None,
    headers=None,
    cookies=None,
    auth=None,
    proxies=None,
    follow_redirects=True,
    cert=None,
    verify=True,
    timeout=5,
    trust_env=True,
):
    return request(
        "GET",
        url,
        params=params,
        headers=headers,
        cookies=cookies,
        auth=auth,
        proxies=proxies,
        follow_redirects=follow_redirects,
        cert=cert,
        verify=verify,
        timeout=timeout,
        trust_env=trust_env,
    )


def options(
    url,
    *,
    params=None,
    headers=None,
    cookies=None,
    auth=None,
    proxies=None,
    follow_redirects=True,
    cert=None,
    verify=True,
    timeout=5,
    trust_env=True,
):
    return request(
        "OPTIONS",
        url,
        params=params,
        headers=headers,
        cookies=cookies,
        auth=auth,
        proxies=proxies,
        follow_redirects=follow_redirects,
        cert=cert,
        verify=verify,
        timeout=timeout,
        trust_env=trust_env,
    )


def head(
    url,
    *,
    params=None,
    headers=None,
    cookies=None,
    auth=None,
    proxies=None,
    follow_redirects=True,
    cert=None,
    verify=True,
    timeout=5,
    trust_env=True,
):
    return request(
        "HEAD",
        url,
        params=params,
        headers=headers,
        cookies=cookies,
        auth=auth,
        proxies=proxies,
        follow_redirects=follow_redirects,
        cert=cert,
        verify=verify,
        timeout=timeout,
        trust_env=trust_env,
    )


def post(
    url,
    *,
    content=None,
    data=None,
    files=None,
    json=None,
    params=None,
    headers=None,
    cookies=None,
    auth=None,
    proxies=None,
    follow_redirects=True,
    cert=None,
    verify=True,
    timeout=5,
    trust_env=True,
):
    return request(
        "POST",
        url,
        content=content,
        data=data,
        files=files,
        json=json,
        params=params,
        headers=headers,
        cookies=cookies,
        auth=auth,
        proxies=proxies,
        follow_redirects=follow_redirects,
        cert=cert,
        verify=verify,
        timeout=timeout,
        trust_env=trust_env,
    )


def put(
    url,
    *,
    content=None,
    data=None,
    files=None,
    json=None,
    params=None,
    headers=None,
    cookies=None,
    auth=None,
    proxies=None,
    follow_redirects=True,
    cert=None,
    verify=True,
    timeout=5,
    trust_env=True,
):
    return request(
        "PUT",
        url,
        content=content,
        data=data,
        files=files,
        json=json,
        params=params,
        headers=headers,
        cookies=cookies,
        auth=auth,
        proxies=proxies,
        follow_redirects=follow_redirects,
        cert=cert,
        verify=verify,
        timeout=timeout,
        trust_env=trust_env,
    )


def patch(
    url,
    *,
    content=None,
    data=None,
    files=None,
    json=None,
    params=None,
    headers=None,
    cookies=None,
    auth=None,
    proxies=None,
    follow_redirects=True,
    cert=None,
    verify=True,
    timeout=5,
    trust_env=True,
):
    return request(
        "PATCH",
        url,
        content=content,
        data=data,
        files=files,
        json=json,
        params=params,
        headers=headers,
        cookies=cookies,
        auth=auth,
        proxies=proxies,
        follow_redirects=follow_redirects,
        cert=cert,
        verify=verify,
        timeout=timeout,
        trust_env=trust_env,
    )


def delete(
    url,
    *,
    params=None,
    headers=None,
    cookies=None,
    auth=None,
    proxies=None,
    follow_redirects=True,
    cert=None,
    verify=True,
    timeout=5,
    trust_env=True,
):
    return request(
        "DELETE",
        url,
        params=params,
        headers=headers,
        cookies=cookies,
        auth=auth,
        proxies=proxies,
        follow_redirects=follow_redirects,
        cert=cert,
        verify=verify,
        timeout=timeout,
        trust_env=trust_env,
    )


@contextmanager
def stream(
    method,
    url,
    *,
    params=None,
    content=None,
    data=None,
    files=None,
    json=None,
    headers=None,
    cookies=None,
    auth=None,
    proxies=None,
    timeout=5,
    follow_redirects=True,
    verify=True,
    cert=None,
    trust_env=True,
):
    with Client(
        cookies=cookies,
        proxies=proxies,
        cert=cert,
        verify=verify,
        timeout=timeout,
        trust_env=trust_env,
    ) as client:
        with client.stream(
            method=method,
            url=url,
            content=content,
            data=data,
            files=files,
            json=json,
            params=params,
            headers=headers,
            auth=auth,
            follow_redirects=follow_redirects,
        ) as response:
            yield response


def wget(
    url,
    method="GET",
    *,
    params=None,
    content=None,
    data=None,
    files=None,
    json=None,
    headers=None,
    cookies=None,
    auth=None,
    proxies=None,
    timeout=5,
    follow_redirects=True,
    verify=True,
    cert=None,
    trust_env=True,
    output=None,
    resume=False
):
    with Client(
        cookies=cookies,
        proxies=proxies,
        cert=cert,
        verify=verify,
        timeout=timeout,
        trust_env=trust_env,
    ) as client:
        return client.wget(
            method=method,
            url=url,
            content=content,
            data=data,
            files=files,
            json=json,
            params=params,
            headers=headers,
            auth=auth,
            follow_redirects=follow_redirects,
            output=output,
            resume=resume
        )



from pyxk.requests.sessions import Session


def request(
    method,
    url,
    *,
    params=None,
    data=None,
    headers=None,
    cookies=None,
    files=None,
    auth=None,
    timeout=5,
    allow_redirects=True,
    proxies=None,
    hooks=None,
    stream=None,
    verify=None,
    cert=None,
    json=None
):
    with Session() as session:
        return session.request(
            method=method,
            url=url,
            params=params,
            data=data,
            headers=headers,
            cookies=cookies,
            files=files,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
            proxies=proxies,
            hooks=hooks,
            stream=stream,
            verify=verify,
            cert=cert,
            json=json,
        )


def get(
    url,
    *,
    params=None,
    data=None,
    headers=None,
    cookies=None,
    files=None,
    auth=None,
    timeout=5,
    allow_redirects=True,
    proxies=None,
    hooks=None,
    stream=None,
    verify=None,
    cert=None,
    json=None
):
    return request(
        method="GET",
        url=url,
        params=params,
        data=data,
        headers=headers,
        cookies=cookies,
        files=files,
        auth=auth,
        timeout=timeout,
        allow_redirects=allow_redirects,
        proxies=proxies,
        hooks=hooks,
        stream=stream,
        verify=verify,
        cert=cert,
        json=json,
    )


def post(
    url,
    *,
    params=None,
    data=None,
    headers=None,
    cookies=None,
    files=None,
    auth=None,
    timeout=5,
    allow_redirects=True,
    proxies=None,
    hooks=None,
    stream=None,
    verify=None,
    cert=None,
    json=None
):
    return request(
        method="POST",
        url=url,
        params=params,
        data=data,
        headers=headers,
        cookies=cookies,
        files=files,
        auth=auth,
        timeout=timeout,
        allow_redirects=allow_redirects,
        proxies=proxies,
        hooks=hooks,
        stream=stream,
        verify=verify,
        cert=cert,
        json=json,
    )


def options(
    url,
    *,
    params=None,
    data=None,
    headers=None,
    cookies=None,
    files=None,
    auth=None,
    timeout=5,
    allow_redirects=True,
    proxies=None,
    hooks=None,
    stream=None,
    verify=None,
    cert=None,
    json=None
):
    return request(
        method="OPTIONS",
        url=url,
        params=params,
        data=data,
        headers=headers,
        cookies=cookies,
        files=files,
        auth=auth,
        timeout=timeout,
        allow_redirects=allow_redirects,
        proxies=proxies,
        hooks=hooks,
        stream=stream,
        verify=verify,
        cert=cert,
        json=json,
    )


def head(
    url,
    *,
    params=None,
    data=None,
    headers=None,
    cookies=None,
    files=None,
    auth=None,
    timeout=5,
    allow_redirects=False,
    proxies=None,
    hooks=None,
    stream=None,
    verify=None,
    cert=None,
    json=None
):
    return request(
        method="HEAD",
        url=url,
        params=params,
        data=data,
        headers=headers,
        cookies=cookies,
        files=files,
        auth=auth,
        timeout=timeout,
        allow_redirects=allow_redirects,
        proxies=proxies,
        hooks=hooks,
        stream=stream,
        verify=verify,
        cert=cert,
        json=json,
    )


def put(
    url,
    *,
    params=None,
    data=None,
    headers=None,
    cookies=None,
    files=None,
    auth=None,
    timeout=5,
    allow_redirects=True,
    proxies=None,
    hooks=None,
    stream=None,
    verify=None,
    cert=None,
    json=None
):
    return request(
        method="PUT",
        url=url,
        params=params,
        data=data,
        headers=headers,
        cookies=cookies,
        files=files,
        auth=auth,
        timeout=timeout,
        allow_redirects=allow_redirects,
        proxies=proxies,
        hooks=hooks,
        stream=stream,
        verify=verify,
        cert=cert,
        json=json,
    )


def patch(
    url,
    *,
    params=None,
    data=None,
    headers=None,
    cookies=None,
    files=None,
    auth=None,
    timeout=5,
    allow_redirects=True,
    proxies=None,
    hooks=None,
    stream=None,
    verify=None,
    cert=None,
    json=None
):
    return request(
        method="PATCH",
        url=url,
        params=params,
        data=data,
        headers=headers,
        cookies=cookies,
        files=files,
        auth=auth,
        timeout=timeout,
        allow_redirects=allow_redirects,
        proxies=proxies,
        hooks=hooks,
        stream=stream,
        verify=verify,
        cert=cert,
        json=json,
    )


def delete(
    url,
    *,
    params=None,
    data=None,
    headers=None,
    cookies=None,
    files=None,
    auth=None,
    timeout=5,
    allow_redirects=True,
    proxies=None,
    hooks=None,
    stream=None,
    verify=None,
    cert=None,
    json=None
):
    return request(
        method="DELETE",
        url=url,
        params=params,
        data=data,
        headers=headers,
        cookies=cookies,
        files=files,
        auth=auth,
        timeout=timeout,
        allow_redirects=allow_redirects,
        proxies=proxies,
        hooks=hooks,
        stream=stream,
        verify=verify,
        cert=cert,
        json=json,
    )


def wget(
    url,
    method="GET",
    *,
    params=None,
    data=None,
    headers=None,
    cookies=None,
    files=None,
    auth=None,
    timeout=5,
    allow_redirects=True,
    proxies=None,
    hooks=None,
    verify=None,
    cert=None,
    json=None,
    stream=True,
    output=None,
    resume=False
):
    """大文件流式响应下载
    
    :params: output: 下载文件路径
    :params: resume: 断点续传(default: False)
    """
    session = Session()
    try:
        response = session.wget(
            method=method,
            url=url,
            params=params,
            data=data,
            headers=headers,
            cookies=cookies,
            files=files,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
            proxies=proxies,
            hooks=hooks,
            verify=verify,
            cert=cert,
            json=json,
            stream=stream,
            output=output,
            resume=resume
        )
    finally:
        session.close()
    return response



from pyxk.lazy_loader import LazyLoader

os = LazyLoader("os", globals())
asyncio = LazyLoader("asyncio", globals())

aiohttp = LazyLoader("aiohttp", globals())
aiofiles = LazyLoader("aiofiles", globals())
progress = LazyLoader("progress", globals(), "rich.progress")

aes = LazyLoader("aes", globals(), "pyxk.aes")
utils = LazyLoader("utils", globals(), "pyxk.utils")


class _AsyncDownload:

    default_timeout = 2 * 60

    def __init__(self, store, limit=8):

        self.store = os.path.join(store, "segments")
        self.limit = limit
        self._temp = store
        self.progress = None
        self.progress_task = None


    def start(self, segments=None, m3u8_keys=None):
        """开启异步下载"""
        if not segments:
            return None

        cipher = {index: aes.Cryptor(**key["cipher"]) for index, key in m3u8_keys.items()}
        os.makedirs(self.store, exist_ok=True)

        with progress.Progress(
            *(
                progress.SpinnerColumn("line"),
                progress.TextColumn("[progress.description]{task.completed}/{task.total}"),
                progress.BarColumn(finished_style="green"),
                progress.TaskProgressColumn("[progress.percentage]{task.percentage:>6.2f}%"),
                progress.TimeElapsedColumn()
            ),
        ) as self.progress:
            self.progress_task = self.progress.add_task(description=None, total=len(segments))
            result = asyncio.run(self.download_manager(segments=segments, cipher=cipher))
        return result


    async def download_manager(self, segments, cipher=None):
        """异步下载管理器"""
        timeout   = aiohttp.ClientTimeout(total=self.default_timeout)
        connector = aiohttp.TCPConnector(limit=self.limit)
        headers   = utils.default_headers()

        async with aiohttp.ClientSession(
            timeout=timeout, connector=connector, headers=headers
        ) as session:
            tasks = [
                self.request(
                    session=session,
                    segment=segment,
                    file=index,
                    cipher=cipher,
                )
                for index, segment in segments.items() # if index < 109
            ]
            result = await asyncio.gather(*tasks)
        return result


    async def request(
        self, session, segment, file, cipher=None
    ):
        """单次异步请求"""
        file = os.path.join(self.store, f"{file}.ts")
        if not os.path.isfile(file) or os.path.getsize(file) == 0:
            content = await self._request(session=session, uri=segment["uri"])
            # 解密 segments
            if cipher and segment["key"] is not None:
                content = cipher[segment["key"]].decrypt(content)
            # 保存 segment
            async with aiofiles.open(file, "wb") as fileobj:
                await fileobj.write(content)

        self.progress.update(self.progress_task, advance=1)
        return file


    @staticmethod
    async def _request(session, uri):
        """获取segment内容"""
        while True:
            try:
                async with session.get(url=uri) as response:
                    # 异常状态码捕获
                    if 403 <= response.status <= 410:
                        raise aiohttp.InvalidURL(
                            f"invalid url:{str(response.url)!r}, status_code: {response.status!r}"
                        )
                    # 重试部分请求
                    if response.status != 200:
                        await asyncio.sleep(1)
                        continue

                    content = await response.content.read()
                    return content
            # 请求超时 重试
            except asyncio.exceptions.TimeoutError:
                await asyncio.sleep(1)
            # 连接错误 重试
            except (
                aiohttp.client_exceptions.ClientOSError,
                aiohttp.client_exceptions.ClientPayloadError,
                aiohttp.client_exceptions.ClientConnectorError,
            ):
                await asyncio.sleep(2)
            # 服务器拒绝连接
            except aiohttp.client_exceptions.ServerDisconnectedError:
                await asyncio.sleep(2)



import requests
from pyxk.lazy_loader import LazyLoader

os = LazyLoader("os", globals())
time = LazyLoader("time", globals())
warnings = LazyLoader("warnings", globals())
urlparse = LazyLoader("urlparse", globals(), "urllib.parse")
utils = LazyLoader("utils", globals(), "pyxk.utils")
console = LazyLoader("console", globals(), "rich.console")
progress = LazyLoader("progress", globals(), "rich.progress")


class Session(requests.Session):
    """requests.Session 重构

    pip install requests
    """

    def __init__(self, *, base_url: str=None):
        super().__init__()
        self.headers["User-Agent"] = utils.UA_ANDROID
        self.base_url = self.set_base_url(base_url)
        self.__console = console.Console()

    def set_base_url(self, url, /):
        """配置 base_url"""
        if url is None:
            return None
        if not self._is_absolute_url(url):
            raise requests.exceptions.InvalidURL(f"{url!r} 不是绝对路径")
        return url

    @staticmethod
    def _is_absolute_url(url, /):
        url = urlparse.urlsplit(url)
        return bool(url.scheme and url.netloc)

    def _build_absolute_url(self, url, /):
        if self.base_url and not self._is_absolute_url(url):
            return urlparse.urljoin(self.base_url, url)
        return url

    def request(self, method, url, params=None, data=None, headers=None, cookies=None, files=None, auth=None, timeout=5, allow_redirects=True, proxies=None, hooks=None, stream=None, verify=None, cert=None, json=None,):
        url, exc_count, exc_max = self._build_absolute_url(url), {}, 10
        while True:
            try:
                with self.__console.status(f"[magenta b]Send Request[/]: {url}"):
                    response = super().request(method=method, url=url, params=params, data=data, headers=headers, cookies=cookies, files=files, auth=auth, timeout=timeout, allow_redirects=allow_redirects, proxies=proxies, hooks=hooks, stream=stream, verify=verify, cert=cert, json=json)
                    break
            except requests.exceptions.Timeout:
                reason = "timeout"
                exc_count.setdefault(reason, 0)
                exc_count[reason] += 1
                if exc_count[reason] >= exc_max:
                    raise
                warnings.warn(f"timeout: {timeout!r}")
                time.sleep(1)
            except requests.exceptions.ConnectionError as exc:
                reason = str(exc.args[0])
                reason_re = ("[Errno 7]", )
                reason_ok = lambda : True in [item in reason for item in reason_re]
                if not reason_ok():
                    raise
                warnings.warn("请检查网络连接是否正常...")
                time.sleep(1)
        return response

    def get(self, url, **kwargs):
        return self.request(method="GET", url=url, **kwargs)

    def options(self, url, **kwargs):
        return self.request(method="OPTIONS", url=url, **kwargs)

    def post(self, url, data=None, json=None, **kwargs):
        return self.request(method="POST", url=url, data=data, json=json, **kwargs)

    def delete(self, url, **kwargs):
        return self.request(method="DELETE", url=url, **kwargs)

    def put(self, url, data=None, **kwargs):
        return self.request(method="PUT", url=url, data=data, **kwargs)

    def patch(self, url, data=None, **kwargs):
        return self.request(method="PATCH", url=url, data=data, **kwargs)

    def head(self, url, **kwargs):
        kwargs.setdefault("allow_redirects", False)
        return self.request(method="HEAD", url=url, **kwargs)

    def wget(self, url, output=None, resume=False, method="GET", *, headers=None, stream=True, **kwargs):
        """大文件流式响应下载

        :params: output: 下载文件路径
        :params: resume: 断点续传(default: False)
        """
        output = self.__set_wget_output(url=url, output=output, headers=headers, **kwargs)
        file_size, file_mode, chunk_size = 0, "wb", 1024
        # 开启文件续传
        if resume:
            if not output:
                warnings.warn("文件续传缺少 output")
            elif os.path.isfile(output):
                file_size, file_mode = os.path.getsize(output), "ab"
                headers = requests.structures.CaseInsensitiveDict(dict(headers or {}))
                headers["Range"] = f"bytes={file_size}-"
        response = self.request(method=method, url=url, headers=headers, stream=stream, **kwargs)
        if not output:
            return response
        content_length = response.headers.get("Content-Length")
        if content_length is not None:
            content_length = int(content_length) + file_size
        with progress.Progress(
            *(
                # progress.SpinnerColumn("line"),
                progress.TextColumn("[progress.description]{task.description}"),
                progress.TaskProgressColumn("[progress.percentage]{task.percentage:>6.2f}%"),
                progress.BarColumn(finished_style="green"),
                progress.DownloadColumn(),
                # progress.TransferSpeedColumn(),
                progress.TimeElapsedColumn()
            ),
            console=self.__console, transient=True
        ) as download_progress:
            description = f"[bold]{os.path.basename(output)}[/]"
            download_task = download_progress.add_task(description=description, total=content_length)
            download_progress.update(download_task, advance=file_size)
            with utils.open(file=output, mode=file_mode) as file_obj:
                for chunk in response.iter_content(chunk_size):
                    file_obj.write(chunk)
                    download_progress.update(download_task, advance=chunk_size)
        return response

    def __set_wget_output(self, url, output, headers, **kwargs):

        if not isinstance(output, str):
            return None
        output = os.path.normpath(os.path.abspath(output))
        if len(os.path.basename(output).rsplit(".", 1)) >= 2:
            return output
        head_response = self.head(url=url, headers=headers, **kwargs)
        content_type = head_response.headers.get("Content-Type")
        if not content_type:
            return output
        content_type = content_type.split(";", 1)[0].strip()
        content_type = content_type.rsplit("/", 1)[-1].strip()
        return output + "." + content_type if content_type else output
